"""Utility functions."""
import json
import logging
import os
import tarfile
from json import JSONDecodeError
from os import makedirs
from os.path import dirname, join
from shutil import copyfile

import click
import requests
from unearthed.cli.errors import bail
from colorama import Back, Fore, Style
from unearthed.cli.auth import auth_handler
from unearthed.cli.config import LAMBDA_URL
from unearthed.core.models.challenge import CHALLENGE_FILENAME, Challenge

logger = logging.getLogger(__name__)


def get_challenge(path=CHALLENGE_FILENAME, id=None):
    """Get a challenge and handle any parsing errors."""
    logger.debug("get_challenge")
    if id is not None:
        try:
            return Challenge.fromApiById(id)
        except Exception as e:
            raise e

    else:
        try:
            return Challenge.fromJsonFile(path)
        except FileNotFoundError as e:
            bail(e, f"No challenge template found in {os.getcwd()}")
        except JSONDecodeError as e:
            bail(e, f"There was a problem decoding the challenge {path}")


def download_from_s3(url, cache_dir=join(os.getcwd(), ".cache")):
    """Download from an S3 signed URL."""
    logger.debug("download_from_s3")

    # Initiate streaming in the archive from url
    if "AWSAccessKeyId" in url:
        logger.debug(f"requesting {url.split('?')[0]}")
    else:
        logger.debug(f"requestion {url}")
    r = requests.get(url, stream=True)

    if r.status_code < 200 or r.status_code > 299:
        logger.debug(f"received HTTP {r.status_code}")
        print(f"There was an error downloading {url.split('?')[0]}.\n")
        print(f"Please contact Unearthed support.")
        exit()

    # determine a name to use in a local cache area
    filename = url.split("?")[0].split("/")[-1]
    cache_filename = join(cache_dir, ".cache", filename)
    logger.debug(f"local cache filename is {cache_filename}")
    makedirs(dirname(cache_filename), exist_ok=True)

    # Initiate the progress tracker of the stream
    total_size = int(r.headers.get("content-length", 0))
    with click.progressbar(length=total_size, label=filename) as bar:
        with open(cache_filename, "wb") as file:
            for data in r.iter_content(1024):
                bar.update(len(data))
                file.write(data)

    return filename


def load_public_data(challenge: Challenge, target_dir=os.getcwd()):
    """Download the public training data."""
    logger.debug("load_public_data")
    filename = download_from_s3(challenge.publicDataUrl, cache_dir=target_dir)
    makedirs(join(target_dir, "data/public"), exist_ok=True)
    copyfile(
        join(target_dir, ".cache", filename), join(target_dir, "data/public", filename)
    )


def load_archive(url, path, name="archive"):
    """Download and unpack an archive from Unearthed Crowd ML S3 bucket."""
    logger.debug("load_archive")

    # Initiate streaming in the archive from url
    if "AWSAccessKeyId" in url:
        logger.debug(f"requesting {url.split('?')[0]}")
    else:
        logger.debug(f"requestion {url}")
    r = requests.get(url, stream=True)

    if r.status_code < 200 or r.status_code > 299:
        logger.debug(f"received HTTP {r.status_code}")
        print(f"There was an error downloading {url.split('?')[0]}.\n")
        print(f"Please contact Unearthed support.")
        exit()

    # TODO this location is not great for large datasets
    tar_path = os.path.expanduser(f"~/.unearthed/downloads/{name}.tar.gz")
    os.makedirs(os.path.dirname(tar_path), exist_ok=True)

    # Initiate the progress tracker of the stream
    total_size = int(r.headers.get("content-length", 0))
    with click.progressbar(length=total_size, label=name) as bar:
        # Save the archive locally
        with open(tar_path, "wb") as tar_file:
            block_size = 1024
            for data in r.iter_content(block_size):
                bar.update(len(data))
                tar_file.write(data)

    # Open and extract the locally saved archive file
    tar = tarfile.open(tar_path)
    os.makedirs(path, exist_ok=True)
    tar.extractall(path=path)


def check_challenge(challenge: Challenge):
    """Check the dataVersion and notifies the user if their template is out-of-date."""
    logger.debug("check_challenge")
    remote_challenge = Challenge.fromApiById(challenge.challengeId)
    if challenge.dataVersion != remote_challenge.dataVersion:
        click.echo(
            Fore.YELLOW
            + "Notice:"
            + Fore.RESET
            + " A new verion of the challenge data or scoring function has been released."
        )
        click.echo(
            "Please update by running: "
            + Fore.LIGHTRED_EX
            + "unearthed reset"
            + Fore.RESET
        )



"""Build command."""
import logging
import os

import click
from unearthed.cli.docker import docker_run
from unearthed.cli.util import get_challenge

logger = logging.getLogger(__name__)


@click.command()
@click.option("--logs/--no-logs", default=True)
def build(logs):
    """Build the submission files using docker and setuptools."""
    logger.debug("build")
    logger.info("Building your submission")

    challenge = get_challenge(".unearthed_challenge.json")

    # the original build sequence uses:
    # - a build step in a Makefile
    #   - (redundantly) installs requirements.txt
    #   - runs setuptools sdist
    #   - renames the resulting package to src.tar.gz
    # - an unpack step in the consuming Makefile
    #   - flattens the python sdist tarball
    #
    # template Makefile
    # requirements:
    #     cat requirements.txt
    #     $(PYTHON_INTERPRETER) -m pip install -U pip setuptools wheel python-dotenv click Sphinx coverage flake8 --quiet || echo "Skipping failed pip install"
    #     # Ensure requirements.txt exists.
    #     cat requirements.txt
    #     $(PYTHON_INTERPRETER) -m pip install -r requirements.txt --quiet || echo "Skipping failed pip install"
    # build:requirements
    # 	rm -rf dist build *.egg-info
    # 	$(PYTHON_INTERPRETER) setup.py sdist bdist_wheel
    # 	cp dist/*.tar.gz dist/src.tar.gz
    #
    # Scoring Makefilecontainer_setup_predict:
    # container_setup_predict:
    # 	# load and unpack model source
    # 	wget -qO- "$(SOURCE_URL)" | tar xz --strip-components=1

    volumes = {os.getcwd(): {"bind": "/tmp/unearthed", "mode": "rw"}}

    # consider setup.py -q to quieten
    cmd = (
        "/bin/bash -c '"
        "cd /tmp/unearthed && python setup.py sdist &&"
        "tar xzf dist/*.tar.gz -C dist --strip-components=1 &&"
        "cd dist && rm -rf *.tar.gz PKG-INFO *.egg-info setup.py setup.cfg pyproject.toml MANIFEST.in &&"
        "tar czf ../src.tar.gz . &&"
        "cd .. && rm -rf dist *.egg-info"
        "'"
    )

    # run setuptools sdist
    result = docker_run(challenge, volumes=volumes, cmd=cmd, logs=logs)
    if result != 1:
        return result
    print(f"build complete")
    return 0



"""Unearthed Configuration."""
from os import environ

if environ.get("DEBUG") or environ.get("UNEARTHED_DEBUG"):
    LAMBDA_URL = "https://crowdml-dev.unearthed.solutions/"
    TRACKER_BASE_URL = "https://development-q5nzhaa-lbjz5xns2dzm2.au.platformsh.site/u"
else:
    LAMBDA_URL = "https://crowdml.unearthed.solutions/"
    TRACKER_BASE_URL = "https://unearthed.solutions/u"



"""Unearthed Cognito."""
import json
import logging
import os

import cognitojwt
from botocore.exceptions import ClientError
from cognitojwt import CognitoJWTException
from cognitojwt.exceptions import CognitoJWTException
from pycognito import Cognito

_TOKEN_PATH = os.path.expanduser("~/.unearthed/tokens.json")

logger = logging.getLogger(__name__)


class UnearthedCognito(Cognito):
    """Authentication handler for Unearthed's Cognito instance."""

    def __init__(self):
        """Use the Unearthed Cognito Pool."""
        super().__init__(
            user_pool_id="us-east-2_yCpz99wwX",
            user_pool_region="us-east-2",
            client_id="1urju2o6sedoa8rkao8d6iino6",
        )
        # logger.debug("__init__")
        # self.__load_tokens()

    def __load_tokens(self):
        """Load saved tokens from user profile/home."""
        try:
            with open(_TOKEN_PATH) as json_file:
                # logger.debug(f"loading tokens from {_TOKEN_PATH}")
                saved_tokens = json.load(json_file)
                self.id_token = saved_tokens["IdToken"]
                self.access_token = saved_tokens["AccessToken"]
                self.refresh_token = saved_tokens["RefreshToken"]
                id_token = self.decoded_id_token(skip_expiry_check=True)
                self.username = id_token["email"]
                # logger.debug(f"tokens loaded for {self.username} from {_TOKEN_PATH}")
        except FileNotFoundError as e:
            logger.debug(f"{_TOKEN_PATH} not found")
        except ClientError as e:
            logger.exception(e)
        except (Exception) as e:
            logger.error(f"failed to read {_TOKEN_PATH}")

    def check_token(self):
        """Check tokens and refresh if required."""
        # logger.debug("check_token")
        self.__load_tokens()
        try:
            refreshed = super().check_token()
            if refreshed:
                self.username = self.decoded_id_token()["email"]
                logger.debug(f"token refreshed for {self.username}")
            return refreshed
        except Exception:
            logger.debug("failure refreshing token, forcing logout")
            self.logout()
            return False

    def __decode(self, token, testmode=False):
        return cognitojwt.decode(
            token,
            self.user_pool_region,
            self.user_pool_id,
            app_client_id=self.client_id,
            testmode=testmode,
        )

    def decoded_id_token(self, skip_expiry_check=False):
        """Decode the Id Token."""
        # logger.debug("decoded_id_token")
        if self.id_token is not None:
            return self.__decode(self.id_token, testmode=skip_expiry_check)
        else:
            return "No Id Token"

    def decoded_access_token(self):
        """Decode the Access Token."""
        logger.debug("decoded_access_token")
        return self.__decode(self.access_token)

    def _save(self):
        logger.debug(f"saving token to {_TOKEN_PATH}")
        os.makedirs(os.path.dirname(_TOKEN_PATH), exist_ok=True)
        tokens = {
            "IdToken": self.id_token,
            "RefreshToken": self.refresh_token,
            "AccessToken": self.access_token,
            # TODO remove email in a later release
            "email": self.username,
            "username": self.username,
        }
        with open(_TOKEN_PATH, "w") as json_file:
            json.dump(tokens, json_file, indent=2)

    def authenticate(self, password):
        """Authenticate the username and password."""
        logger.debug("authenticate")
        super().authenticate(password)
        self._save()

    def logout(self):
        """Log out the current session and remove token caches."""
        logger.debug("logout")
        if os.path.exists(_TOKEN_PATH):
            logger.debug(f"removing {_TOKEN_PATH}")
            os.remove(_TOKEN_PATH)

        if self.access_token is not None:
            return super().logout()

    def is_logged_in(self):
        """Is the current session logged in."""
        self.check_token()
        return os.path.exists(_TOKEN_PATH)



"""Unearthed CLI Authentiation Commands."""
import os
import json
import logging
import sys

import click
from botocore.exceptions import ClientError
from unearthed.cli.auth import authorize, cognito_session

logger = logging.getLogger(__name__)


def _validate_username(ctx, param, value):
    """Prompt for username if not found in environment."""
    if value is None:
        return click.prompt('Unearthed Login Email',
                            default=cognito_session.username)
    else:
        print('Found UNEARTHED_USERNAME environment variable')
        return value


def _validate_password(ctx, param, value):
    """Prompt for password if not found in environment."""
    if value is None:
        return click.prompt('Unearthed Login Password',
                            hide_input=True)
    else:
        print('Found UNEARTHED_PASSWORD environment variable')
        return value


@click.command()
@click.option('--username', callback=_validate_username,
              envvar='UNEARTHED_USERNAME')
@click.option('--password', callback=_validate_password,
              envvar='UNEARTHED_PASSWORD')
def login(username, password):
    """Login to Unearthed."""
    logger.debug('login')

    # force a token refresh if possible
    cognito_session.check_token()
    if username == cognito_session.username:
        return print(f'Already logged in to Unearthed as {username}')

    try:
        cognito_session.username = username
        cognito_session.authenticate(password)
        print(f'Successfully logged in to Unearthed as {username}!')
    except ClientError as e:
        print(e.response['message'])


@click.command()
def logout():
    """Logout of Unearthed."""
    logger.debug('logout')
    cognito_session.logout()
    print('You are now logged out of Unearthed.')


@click.group()
def auth():
    """Manage authentication."""
    pass


@auth.command('whoami')
def whoami():
    """Display current username."""
    logger.debug('whoami')
    cognito_session.check_token()
    print(cognito_session.username)


@auth.group()
def tokens():
    """Display authentication tokens."""
    pass


@tokens.command()
def access():
    """Display the decoded Access Token."""
    logger.debug('access')
    print(json.dumps(cognito_session.decoded_access_token(), indent=2))


@tokens.command()
def id():
    """Display the decoded Id Token."""
    logger.debug('id')
    print(json.dumps(cognito_session.decoded_id_token(), indent=2))



"""Unearthed Submissions."""
import logging
import os
import webbrowser
from json.decoder import JSONDecodeError

import click
import docker
import requests
from unearthed.cli.errors import bail
from unearthed.cli.auth import authorize
from unearthed.cli.commands.build import build
from unearthed.cli.config import TRACKER_BASE_URL
from unearthed.cli.docker import docker_run
from unearthed.cli.util import check_challenge, get_challenge
from unearthed.core.models.submission import Submission

logger = logging.getLogger("unearthed." + __name__)


@click.command()
@click.option("--tracker/--no-tracker", default=True)
@click.option("--logs/--no-logs", default=False)
@click.option("--check/--no-check", default=True)
@click.pass_context
@authorize
def submit(ctx, tracker, logs, check):
    """Create a new submission and upload source code to Unearthed."""
    logger.debug("submit")
    challenge = get_challenge()

    # Check that the challenge is up-to-date
    if check:
        check_challenge(challenge)

    # Bundle the source code
    click.echo("Building your submission...")
    exit_code = ctx.invoke(build, logs=logs)
    if exit_code != 0:
        return click.echo("Build Failed. Try adding --logs to see the build logs.")

    # Create the submission
    click.echo("Creating submission...")
    try:
        submission = Submission.fromApi(challenge.challengeId)
        logger.debug(f"submission id is {submission.submissionId}")
    except JSONDecodeError as e:
        bail(e, "There was a problem contacting the server. Please try again.")
    except Exception as e:
        bail(e, "There was a problem contacting the server.")

    # make sure the source upload URL is present
    if submission.sourceUploadUrl == "":
        return click.echo("Something went wrong getting upload URL")

    # Upload the config file to S3
    with open("submission.yml", "rb") as data:
        response = requests.put(url=submission.configUploadUrl, data=data, timeout=60)
        response.raise_for_status()

    # Upload the source code to S3
    click.echo("Sending submission...")
    logger.debug(f"sending src.tar.gz to {submission.sourceUploadUrl.split('?')[0]}")
    with open("src.tar.gz", "rb") as data:
        response = requests.put(url=submission.sourceUploadUrl, data=data, timeout=60)
        response.raise_for_status()

    # Open the tracker
    click.echo("Submission complete!")
    tracker_url = (
        f"{TRACKER_BASE_URL}/crowdml/tracker/"
        f"{submission.innovatorPortalSubmissionRef}"
    )
    click.echo(f"You can track the progress of your submission at: {tracker_url}")
    if tracker:
        webbrowser.open_new_tab(tracker_url)



"""Docker wrapper."""
import json
import logging
import os
import sys


from unearthed.cli.errors import bail
from tqdm import tqdm

import docker
from docker.errors import ImageNotFound

logger = logging.getLogger(__name__)

client = docker.from_env()
# image_repo = 'qc5i1je5hh.execute-api.us-east-2.amazonaws.com'


def image_exists(image_name: str) -> bool:
    """Check a docker image exists."""
    logger.debug("image_exists")
    try:
        client.images.get(image_name)
        return True
    except ImageNotFound:
        return False


def pull_image(image_name):
    """Pull a docker image and display a status bar."""
    bars = {}
    pos = 0

    for update in client.api.pull(image_name, stream=True, decode=True):
        status = update.get("status", "")
        bar = update.get("id", "")
        progress = update.get("progress", "")
        if bar not in bars:
            bars[bar] = tqdm(position=pos, initial=1, leave=True)
            bars[bar].display("")
            pos += 1
        bars[bar].display(" ".join([bar, status, progress]))

    [bars[bar].close() for bar in bars]


# Abstraction command for running a command in the container
def docker_run(chall, cmd="", env=[], volumes={}, logs=True, pull=False):
    """Run a command in a Docker image."""
    logger.debug("docker_run")
    # TODO probably need to pass the image in from a config
    # it is in the mongo db, not sure if this is the right place for it
    docker_image = "python:3.8-slim"

    print(f"docker run -t {docker_image}")
    try:
        # only pull if image does not exist or flag is set
        if pull or not image_exists(docker_image):
            pull_image(docker_image)
    except Exception as e:
        bail(
            e,
            "There was a problem fetching the Docker image.\n"
            "Please make sure you have Docker installed.",
        )

    print("Running command on image " + docker_image)

    # Start the command execution
    container = client.containers.run(
        image=docker_image,
        command=cmd,
        # command="bash -c '{}'".format(cmd),
        auto_remove=True,
        volumes=volumes,
        environment=env,
        network_mode="host",
        detach=True,
    )
    # If enabled, output the logs
    logs = True
    if logs:
        log_stream = container.logs(stream=True)
        for line in log_stream:
            print(line.decode("utf-8"), end="")
    # Wait until the container is finished executing
    res = container.wait()
    # Return the exit code of the command
    return res["StatusCode"]
